@@ -22,10 +22,12 @@
 
 #include "backend/x11.h"
 #include "util/time.h"
+#include "types/wlr_output.h"
 
 static const uint32_t SUPPORTED_OUTPUT_STATE =
 	WLR_OUTPUT_STATE_BACKEND_OPTIONAL |
 	WLR_OUTPUT_STATE_BUFFER |
+	WLR_OUTPUT_STATE_ENABLED |
 	WLR_OUTPUT_STATE_MODE |
 	WLR_OUTPUT_STATE_ADAPTIVE_SYNC_ENABLED;
 
@@ -53,7 +55,8 @@ static void parse_xcb_setup(struct wlr_output *output,
 static struct wlr_x11_output *get_x11_output_from_output(
 		struct wlr_output *wlr_output) {
 	assert(wlr_output_is_x11(wlr_output));
-	return (struct wlr_x11_output *)wlr_output;
+	struct wlr_x11_output *output = wl_container_of(wlr_output, output, wlr_output);
+	return output;
 }
 
 static bool output_set_custom_mode(struct wlr_output *wlr_output,
@@ -74,6 +77,9 @@ static bool output_set_custom_mode(struct wlr_output *wlr_output,
 		return false;
 	}
 
+	// Move the pointer to its new location
+	update_x11_pointer_position(output, output->x11->time);
+
 	return true;
 }
 
@@ -122,12 +128,18 @@ static bool output_test(struct wlr_output *wlr_output,
 	assert(wlr_output->adaptive_sync_status == WLR_OUTPUT_ADAPTIVE_SYNC_ENABLED);
 	if (state->committed & WLR_OUTPUT_STATE_ADAPTIVE_SYNC_ENABLED) {
 		if (!state->adaptive_sync_enabled) {
+			wlr_log(WLR_DEBUG, "Disabling adaptive sync is not supported");
 			return false;
 		}
 	}
 
 	if (state->committed & WLR_OUTPUT_STATE_MODE) {
 		assert(state->mode_type == WLR_OUTPUT_STATE_MODE_CUSTOM);
+
+		if (state->custom_mode.refresh != 0) {
+			wlr_log(WLR_DEBUG, "Refresh rates are not supported");
+			return false;
+		}
 	}
 
 	return true;
@@ -244,7 +256,7 @@ static struct wlr_x11_buffer *create_x11_buffer(struct wlr_x11_output *output,
 		return NULL;
 	}
 
-	struct wlr_x11_buffer *buffer = calloc(1, sizeof(struct wlr_x11_buffer));
+	struct wlr_x11_buffer *buffer = calloc(1, sizeof(*buffer));
 	if (!buffer) {
 		xcb_free_pixmap(x11->xcb, pixmap);
 		return NULL;
@@ -288,11 +300,10 @@ static bool output_commit_buffer(struct wlr_x11_output *output,
 
 	xcb_xfixes_region_t region = XCB_NONE;
 	if (state->committed & WLR_OUTPUT_STATE_DAMAGE) {
-		pixman_region32_union(&output->exposed, &output->exposed,
-			(pixman_region32_t *) &state->damage);
+		pixman_region32_union(&output->exposed, &output->exposed, &state->damage);
 
 		int rects_len = 0;
-		pixman_box32_t *rects = pixman_region32_rectangles(&output->exposed, &rects_len);
+		const pixman_box32_t *rects = pixman_region32_rectangles(&output->exposed, &rects_len);
 
 		xcb_rectangle_t *xcb_rects = calloc(rects_len, sizeof(xcb_rectangle_t));
 		if (!xcb_rects) {
@@ -300,7 +311,7 @@ static bool output_commit_buffer(struct wlr_x11_output *output,
 		}
 
 		for (int i = 0; i < rects_len; i++) {
-			pixman_box32_t *box = &rects[i];
+			const pixman_box32_t *box = &rects[i];
 			xcb_rects[i] = (struct xcb_rectangle_t){
 				.x = box->x1,
 				.y = box->y1,
@@ -344,6 +355,14 @@ static bool output_commit(struct wlr_output *wlr_output,
 		return false;
 	}
 
+	if (state->committed & WLR_OUTPUT_STATE_ENABLED) {
+		if (state->enabled) {
+			xcb_map_window(x11->xcb, output->win);
+		} else {
+			xcb_unmap_window(x11->xcb, output->win);
+		}
+	}
+
 	if (state->committed & WLR_OUTPUT_STATE_MODE) {
 		if (!output_set_custom_mode(wlr_output,
 				state->custom_mode.width,
@@ -357,6 +376,10 @@ static bool output_commit(struct wlr_output *wlr_output,
 		if (!output_commit_buffer(output, state)) {
 			return false;
 		}
+	} else if (output_pending_enabled(wlr_output, state)) {
+		uint32_t serial = output->wlr_output.commit_seq;
+		uint64_t target_msc = output->last_msc ? output->last_msc + 1 : 0;
+		xcb_present_notify_msc(x11->xcb, output->win, serial, target_msc, 0, 0);
 	}
 
 	xcb_flush(x11->xcb);
@@ -512,7 +535,7 @@ struct wlr_output *wlr_x11_output_create(struct wlr_backend *backend) {
 		return NULL;
 	}
 
-	struct wlr_x11_output *output = calloc(1, sizeof(struct wlr_x11_output));
+	struct wlr_x11_output *output = calloc(1, sizeof(*output));
 	if (output == NULL) {
 		return NULL;
 	}
@@ -521,9 +544,13 @@ struct wlr_output *wlr_x11_output_create(struct wlr_backend *backend) {
 	pixman_region32_init(&output->exposed);
 
 	struct wlr_output *wlr_output = &output->wlr_output;
-	wlr_output_init(wlr_output, &x11->backend, &output_impl, x11->wl_display);
 
-	wlr_output_update_custom_mode(wlr_output, 1024, 768, 0);
+	struct wlr_output_state state;
+	wlr_output_state_init(&state);
+	wlr_output_state_set_custom_mode(&state, 1024, 768, 0);
+
+	wlr_output_init(wlr_output, &x11->backend, &output_impl, x11->wl_display, &state);
+	wlr_output_state_finish(&state);
 
 	size_t output_num = ++last_output_num;
 
@@ -586,13 +613,10 @@ struct wlr_output *wlr_x11_output_create(struct wlr_backend *backend) {
 
 	wlr_x11_output_set_title(wlr_output, NULL);
 
-	xcb_map_window(x11->xcb, output->win);
 	xcb_flush(x11->xcb);
 
 	wl_list_insert(&x11->outputs, &output->link);
 
-	wlr_output_update_enabled(wlr_output, true);
-
 	wlr_pointer_init(&output->pointer, &x11_pointer_impl, "x11-pointer");
 	output->pointer.output_name = strdup(wlr_output->name);
 
@@ -604,15 +628,11 @@ struct wlr_output *wlr_x11_output_create(struct wlr_backend *backend) {
 	wl_signal_emit_mutable(&x11->backend.events.new_input, &output->pointer.base);
 	wl_signal_emit_mutable(&x11->backend.events.new_input, &output->touch.base);
 
-	// Start the rendering loop by requesting the compositor to render a frame
-	wlr_output_schedule_frame(wlr_output);
-
 	return wlr_output;
 }
 
 void handle_x11_configure_notify(struct wlr_x11_output *output,
 		xcb_configure_notify_event_t *ev) {
-	// ignore events that set an invalid size:
 	if (ev->width == 0 || ev->height == 0) {
 		wlr_log(WLR_DEBUG,
 			"Ignoring X11 configure event for height=%d, width=%d",
@@ -620,11 +640,11 @@ void handle_x11_configure_notify(struct wlr_x11_output *output,
 		return;
 	}
 
-	wlr_output_update_custom_mode(&output->wlr_output, ev->width,
-		ev->height, 0);
-
-	// Move the pointer to its new location
-	update_x11_pointer_position(output, output->x11->time);
+	struct wlr_output_state state;
+	wlr_output_state_init(&state);
+	wlr_output_state_set_custom_mode(&state, ev->width, ev->height, 0);
+	wlr_output_send_request_state(&output->wlr_output, &state);
+	wlr_output_state_finish(&state);
 }
 
 bool wlr_output_is_x11(struct wlr_output *wlr_output) {
