@@ -2,40 +2,18 @@
 #include <float.h>
 #include <limits.h>
 #include <stdlib.h>
-#include <wlr/types/wlr_box.h>
 #include <wlr/types/wlr_output_layout.h>
 #include <wlr/types/wlr_output.h>
+#include <wlr/util/box.h>
 #include <wlr/util/log.h>
-#include "util/signal.h"
 
-struct wlr_output_layout_state {
-	struct wlr_box _box; // should never be read directly, use the getter
-};
-
-struct wlr_output_layout_output_state {
-	struct wlr_output_layout *layout;
-	struct wlr_output_layout_output *l_output;
-
-	struct wlr_box _box; // should never be read directly, use the getter
-	bool auto_configured;
-
-	struct wl_listener mode;
-	struct wl_listener scale;
-	struct wl_listener transform;
-	struct wl_listener output_destroy;
-};
+static const struct wlr_addon_interface addon_impl;
 
 struct wlr_output_layout *wlr_output_layout_create(void) {
-	struct wlr_output_layout *layout =
-		calloc(1, sizeof(struct wlr_output_layout));
+	struct wlr_output_layout *layout = calloc(1, sizeof(*layout));
 	if (layout == NULL) {
 		return NULL;
 	}
-	layout->state = calloc(1, sizeof(struct wlr_output_layout_state));
-	if (layout->state == NULL) {
-		free(layout);
-		return NULL;
-	}
 	wl_list_init(&layout->outputs);
 
 	wl_signal_init(&layout->events.add);
@@ -47,14 +25,11 @@ struct wlr_output_layout *wlr_output_layout_create(void) {
 
 static void output_layout_output_destroy(
 		struct wlr_output_layout_output *l_output) {
-	wlr_signal_emit_safe(&l_output->events.destroy, l_output);
+	wl_signal_emit_mutable(&l_output->events.destroy, l_output);
 	wlr_output_destroy_global(l_output->output);
-	wl_list_remove(&l_output->state->mode.link);
-	wl_list_remove(&l_output->state->scale.link);
-	wl_list_remove(&l_output->state->transform.link);
-	wl_list_remove(&l_output->state->output_destroy.link);
+	wl_list_remove(&l_output->commit.link);
 	wl_list_remove(&l_output->link);
-	free(l_output->state);
+	wlr_addon_finish(&l_output->addon);
 	free(l_output);
 }
 
@@ -63,26 +38,23 @@ void wlr_output_layout_destroy(struct wlr_output_layout *layout) {
 		return;
 	}
 
-	wlr_signal_emit_safe(&layout->events.destroy, layout);
+	wl_signal_emit_mutable(&layout->events.destroy, layout);
 
 	struct wlr_output_layout_output *l_output, *temp;
 	wl_list_for_each_safe(l_output, temp, &layout->outputs, link) {
 		output_layout_output_destroy(l_output);
 	}
 
-	free(layout->state);
 	free(layout);
 }
 
-static struct wlr_box *output_layout_output_get_box(
-		struct wlr_output_layout_output *l_output) {
-	l_output->state->_box.x = l_output->x;
-	l_output->state->_box.y = l_output->y;
-	int width, height;
-	wlr_output_effective_resolution(l_output->output, &width, &height);
-	l_output->state->_box.width = width;
-	l_output->state->_box.height = height;
-	return &l_output->state->_box;
+static void output_layout_output_get_box(
+		struct wlr_output_layout_output *l_output,
+		struct wlr_box *box) {
+	box->x = l_output->x;
+	box->y = l_output->y;
+	wlr_output_effective_resolution(l_output->output,
+		&box->width, &box->height);
 }
 
 /**
@@ -99,15 +71,17 @@ static void output_layout_reconfigure(struct wlr_output_layout *layout) {
 	// find the rightmost x coordinate occupied by a manually configured output
 	// in the layout
 	struct wlr_output_layout_output *l_output;
+	struct wlr_box output_box;
+
 	wl_list_for_each(l_output, &layout->outputs, link) {
-		if (l_output->state->auto_configured) {
+		if (l_output->auto_configured) {
 			continue;
 		}
 
-		struct wlr_box *box = output_layout_output_get_box(l_output);
-		if (box->x + box->width > max_x) {
-			max_x = box->x + box->width;
-			max_x_y = box->y;
+		output_layout_output_get_box(l_output, &output_box);
+		if (output_box.x + output_box.width > max_x) {
+			max_x = output_box.x + output_box.width;
+			max_x_y = output_box.y;
 		}
 	}
 
@@ -118,16 +92,16 @@ static void output_layout_reconfigure(struct wlr_output_layout *layout) {
 	}
 
 	wl_list_for_each(l_output, &layout->outputs, link) {
-		if (!l_output->state->auto_configured) {
+		if (!l_output->auto_configured) {
 			continue;
 		}
-		struct wlr_box *box = output_layout_output_get_box(l_output);
+		output_layout_output_get_box(l_output, &output_box);
 		l_output->x = max_x;
 		l_output->y = max_x_y;
-		max_x += box->width;
+		max_x += output_box.width;
 	}
 
-	wlr_signal_emit_safe(&layout->events.change, layout);
+	wl_signal_emit_mutable(&layout->events.change, layout);
 }
 
 static void output_update_global(struct wlr_output *output) {
@@ -139,96 +113,113 @@ static void output_update_global(struct wlr_output *output) {
 	}
 }
 
-static void handle_output_mode(struct wl_listener *listener, void *data) {
-	struct wlr_output_layout_output_state *state =
-		wl_container_of(listener, state, mode);
-	output_layout_reconfigure(state->layout);
-	output_update_global(state->l_output->output);
+static void handle_output_commit(struct wl_listener *listener, void *data) {
+	struct wlr_output_layout_output *l_output =
+		wl_container_of(listener, l_output, commit);
+	struct wlr_output_event_commit *event = data;
+
+	if (event->state->committed & (WLR_OUTPUT_STATE_SCALE |
+			WLR_OUTPUT_STATE_TRANSFORM |
+			WLR_OUTPUT_STATE_MODE)) {
+		output_layout_reconfigure(l_output->layout);
+		output_update_global(l_output->output);
+	}
 }
 
-static void handle_output_scale(struct wl_listener *listener, void *data) {
-	struct wlr_output_layout_output_state *state =
-		wl_container_of(listener, state, scale);
-	output_layout_reconfigure(state->layout);
-}
-
-static void handle_output_transform(struct wl_listener *listener, void *data) {
-	struct wlr_output_layout_output_state *state =
-		wl_container_of(listener, state, transform);
-	output_layout_reconfigure(state->layout);
-}
-
-static void handle_output_destroy(struct wl_listener *listener, void *data) {
-	struct wlr_output_layout_output_state *state =
-		wl_container_of(listener, state, output_destroy);
-	struct wlr_output_layout *layout = state->layout;
-	output_layout_output_destroy(state->l_output);
+static void addon_destroy(struct wlr_addon *addon) {
+	assert(addon->impl == &addon_impl);
+	struct wlr_output_layout_output *l_output =
+		wl_container_of(addon, l_output, addon);
+	struct wlr_output_layout *layout = l_output->layout;
+	output_layout_output_destroy(l_output);
 	output_layout_reconfigure(layout);
 }
 
+static const struct wlr_addon_interface addon_impl = {
+	.name = "wlr_output_layout_output",
+	.destroy = addon_destroy,
+};
+
 static struct wlr_output_layout_output *output_layout_output_create(
 		struct wlr_output_layout *layout, struct wlr_output *output) {
-	struct wlr_output_layout_output *l_output =
-		calloc(1, sizeof(struct wlr_output_layout_output));
+	struct wlr_output_layout_output *l_output = calloc(1, sizeof(*l_output));
 	if (l_output == NULL) {
 		return NULL;
 	}
-	l_output->state = calloc(1, sizeof(struct wlr_output_layout_output_state));
-	if (l_output->state == NULL) {
-		free(l_output);
-		return NULL;
-	}
-	l_output->state->l_output = l_output;
-	l_output->state->layout = layout;
+	l_output->layout = layout;
 	l_output->output = output;
 	wl_signal_init(&l_output->events.destroy);
-	wl_list_insert(&layout->outputs, &l_output->link);
 
-	wl_signal_add(&output->events.mode, &l_output->state->mode);
-	l_output->state->mode.notify = handle_output_mode;
-	wl_signal_add(&output->events.scale, &l_output->state->scale);
-	l_output->state->scale.notify = handle_output_scale;
-	wl_signal_add(&output->events.transform, &l_output->state->transform);
-	l_output->state->transform.notify = handle_output_transform;
-	wl_signal_add(&output->events.destroy, &l_output->state->output_destroy);
-	l_output->state->output_destroy.notify = handle_output_destroy;
+	/*
+	 * Insert at the end of the list so that auto-configuring the
+	 * new output doesn't change the layout of other outputs
+	 */
+	wl_list_insert(layout->outputs.prev, &l_output->link);
+
+	wl_signal_add(&output->events.commit, &l_output->commit);
+	l_output->commit.notify = handle_output_commit;
+
+	wlr_addon_init(&l_output->addon, &output->addons, layout, &addon_impl);
 
 	return l_output;
 }
 
-void wlr_output_layout_add(struct wlr_output_layout *layout,
-		struct wlr_output *output, int lx, int ly) {
+static struct wlr_output_layout_output *output_layout_add(struct wlr_output_layout *layout,
+		struct wlr_output *output, int lx, int ly,
+		bool auto_configured) {
 	struct wlr_output_layout_output *l_output =
 		wlr_output_layout_get(layout, output);
 	bool is_new = l_output == NULL;
-	if (!l_output) {
+	if (is_new) {
 		l_output = output_layout_output_create(layout, output);
-		if (!l_output) {
-			wlr_log(WLR_ERROR, "Failed to create wlr_output_layout_output");
-			return;
+		if (l_output == NULL) {
+			return NULL;
 		}
 	}
 
 	l_output->x = lx;
 	l_output->y = ly;
-	l_output->state->auto_configured = false;
+	l_output->auto_configured = auto_configured;
+
 	output_layout_reconfigure(layout);
 	output_update_global(output);
 
 	if (is_new) {
-		wlr_signal_emit_safe(&layout->events.add, l_output);
+		wl_signal_emit_mutable(&layout->events.add, l_output);
+	}
+
+	return l_output;
+}
+
+struct wlr_output_layout_output *wlr_output_layout_add(struct wlr_output_layout *layout,
+		struct wlr_output *output, int lx, int ly) {
+	return output_layout_add(layout, output, lx, ly, false);
+}
+
+struct wlr_output_layout_output *wlr_output_layout_add_auto(struct wlr_output_layout *layout,
+		struct wlr_output *output) {
+	return output_layout_add(layout, output, 0, 0, true);
+}
+
+void wlr_output_layout_remove(struct wlr_output_layout *layout,
+		struct wlr_output *output) {
+	struct wlr_output_layout_output *l_output =
+		wlr_output_layout_get(layout, output);
+	if (l_output != NULL) {
+		output_layout_output_destroy(l_output);
+		output_layout_reconfigure(layout);
 	}
 }
 
 struct wlr_output_layout_output *wlr_output_layout_get(
 		struct wlr_output_layout *layout, struct wlr_output *reference) {
-	struct wlr_output_layout_output *l_output;
-	wl_list_for_each(l_output, &layout->outputs, link) {
-		if (l_output->output == reference) {
-			return l_output;
-		}
+	struct wlr_output_layout_output *l_output = NULL;
+	struct wlr_addon *addon =
+		wlr_addon_find(&reference->addons, layout, &addon_impl);
+	if (addon) {
+		l_output = wl_container_of(addon, l_output, addon);
 	}
-	return NULL;
+	return l_output;
 }
 
 bool wlr_output_layout_contains_point(struct wlr_output_layout *layout,
@@ -236,8 +227,12 @@ bool wlr_output_layout_contains_point(struct wlr_output_layout *layout,
 	if (reference) {
 		struct wlr_output_layout_output *l_output =
 			wlr_output_layout_get(layout, reference);
-		struct wlr_box *box = output_layout_output_get_box(l_output);
-		return wlr_box_contains_point(box, lx, ly);
+		if (!l_output) {
+			return false;
+		}
+		struct wlr_box output_box;
+		output_layout_output_get_box(l_output, &output_box);
+		return wlr_box_contains_point(&output_box, lx, ly);
 	} else {
 		return !!wlr_output_layout_output_at(layout, lx, ly);
 	}
@@ -250,9 +245,9 @@ bool wlr_output_layout_intersects(struct wlr_output_layout *layout,
 	if (reference == NULL) {
 		struct wlr_output_layout_output *l_output;
 		wl_list_for_each(l_output, &layout->outputs, link) {
-			struct wlr_box *output_box =
-				output_layout_output_get_box(l_output);
-			if (wlr_box_intersection(&out_box, output_box, target_lbox)) {
+			struct wlr_box output_box;
+			output_layout_output_get_box(l_output, &output_box);
+			if (wlr_box_intersection(&out_box, &output_box, target_lbox)) {
 				return true;
 			}
 		}
@@ -264,8 +259,9 @@ bool wlr_output_layout_intersects(struct wlr_output_layout *layout,
 			return false;
 		}
 
-		struct wlr_box *output_box = output_layout_output_get_box(l_output);
-		return wlr_box_intersection(&out_box, output_box, target_lbox);
+		struct wlr_box output_box;
+		output_layout_output_get_box(l_output, &output_box);
+		return wlr_box_intersection(&out_box, &output_box, target_lbox);
 	}
 }
 
@@ -273,38 +269,15 @@ struct wlr_output *wlr_output_layout_output_at(struct wlr_output_layout *layout,
 		double lx, double ly) {
 	struct wlr_output_layout_output *l_output;
 	wl_list_for_each(l_output, &layout->outputs, link) {
-		struct wlr_box *box = output_layout_output_get_box(l_output);
-		if (wlr_box_contains_point(box, lx, ly)) {
+		struct wlr_box output_box;
+		output_layout_output_get_box(l_output, &output_box);
+		if (wlr_box_contains_point(&output_box, lx, ly)) {
 			return l_output->output;
 		}
 	}
 	return NULL;
 }
 
-void wlr_output_layout_move(struct wlr_output_layout *layout,
-		struct wlr_output *output, int lx, int ly) {
-	struct wlr_output_layout_output *l_output =
-		wlr_output_layout_get(layout, output);
-	if (l_output) {
-		l_output->x = lx;
-		l_output->y = ly;
-		l_output->state->auto_configured = false;
-		output_layout_reconfigure(layout);
-	} else {
-		wlr_log(WLR_ERROR, "output not found in this layout: %s", output->name);
-	}
-}
-
-void wlr_output_layout_remove(struct wlr_output_layout *layout,
-		struct wlr_output *output) {
-	struct wlr_output_layout_output *l_output =
-		wlr_output_layout_get(layout, output);
-	if (l_output) {
-		output_layout_output_destroy(l_output);
-		output_layout_reconfigure(layout);
-	}
-}
-
 void wlr_output_layout_output_coords(struct wlr_output_layout *layout,
 		struct wlr_output *reference, double *lx, double *ly) {
 	assert(layout && reference);
@@ -328,7 +301,7 @@ void wlr_output_layout_closest_point(struct wlr_output_layout *layout,
 		return;
 	}
 
-	double min_x = 0, min_y = 0, min_distance = DBL_MAX;
+	double min_x = lx, min_y = ly, min_distance = DBL_MAX;
 	struct wlr_output_layout_output *l_output;
 	wl_list_for_each(l_output, &layout->outputs, link) {
 		if (reference != NULL && reference != l_output->output) {
@@ -336,8 +309,9 @@ void wlr_output_layout_closest_point(struct wlr_output_layout *layout,
 		}
 
 		double output_x, output_y, output_distance;
-		struct wlr_box *box = output_layout_output_get_box(l_output);
-		wlr_box_closest_point(box, lx, ly, &output_x, &output_y);
+		struct wlr_box output_box;
+		output_layout_output_get_box(l_output, &output_box);
+		wlr_box_closest_point(&output_box, lx, ly, &output_x, &output_y);
 
 		// calculate squared distance suitable for comparison
 		output_distance =
@@ -362,17 +336,17 @@ void wlr_output_layout_closest_point(struct wlr_output_layout *layout,
 	}
 }
 
-struct wlr_box *wlr_output_layout_get_box(
-		struct wlr_output_layout *layout, struct wlr_output *reference) {
+void wlr_output_layout_get_box(struct wlr_output_layout *layout,
+		struct wlr_output *reference, struct wlr_box *dest_box) {
+	*dest_box = (struct wlr_box){0};
+
 	struct wlr_output_layout_output *l_output;
 	if (reference) {
 		// output extents
 		l_output = wlr_output_layout_get(layout, reference);
 
 		if (l_output) {
-			return output_layout_output_get_box(l_output);
-		} else {
-			return NULL;
+			output_layout_output_get_box(l_output, dest_box);
 		}
 	} else {
 		// layout extents
@@ -381,52 +355,27 @@ struct wlr_box *wlr_output_layout_get_box(
 			min_x = min_y = INT_MAX;
 			max_x = max_y = INT_MIN;
 			wl_list_for_each(l_output, &layout->outputs, link) {
-				struct wlr_box *box = output_layout_output_get_box(l_output);
-				if (box->x < min_x) {
-					min_x = box->x;
+				struct wlr_box output_box;
+				output_layout_output_get_box(l_output, &output_box);
+				if (output_box.x < min_x) {
+					min_x = output_box.x;
 				}
-				if (box->y < min_y) {
-					min_y = box->y;
+				if (output_box.y < min_y) {
+					min_y = output_box.y;
 				}
-				if (box->x + box->width > max_x) {
-					max_x = box->x + box->width;
+				if (output_box.x + output_box.width > max_x) {
+					max_x = output_box.x + output_box.width;
 				}
-				if (box->y + box->height > max_y) {
-					max_y = box->y + box->height;
+				if (output_box.y + output_box.height > max_y) {
+					max_y = output_box.y + output_box.height;
 				}
 			}
 		}
 
-		layout->state->_box.x = min_x;
-		layout->state->_box.y = min_y;
-		layout->state->_box.width = max_x - min_x;
-		layout->state->_box.height = max_y - min_y;
-
-		return &layout->state->_box;
-	}
-
-	// not reached
-}
-
-void wlr_output_layout_add_auto(struct wlr_output_layout *layout,
-		struct wlr_output *output) {
-	struct wlr_output_layout_output *l_output =
-		wlr_output_layout_get(layout, output);
-	bool is_new = l_output == NULL;
-	if (!l_output) {
-		l_output = output_layout_output_create(layout, output);
-		if (!l_output) {
-			wlr_log(WLR_ERROR, "Failed to create wlr_output_layout_output");
-			return;
-		}
-	}
-
-	l_output->state->auto_configured = true;
-	output_layout_reconfigure(layout);
-	output_update_global(output);
-
-	if (is_new) {
-		wlr_signal_emit_safe(&layout->events.add, l_output);
+		dest_box->x = min_x;
+		dest_box->y = min_y;
+		dest_box->width = max_x - min_x;
+		dest_box->height = max_y - min_y;
 	}
 }
 
@@ -436,9 +385,10 @@ struct wlr_output *wlr_output_layout_get_center_output(
 		return NULL;
 	}
 
-	struct wlr_box *extents = wlr_output_layout_get_box(layout, NULL);
-	double center_x = extents->width / 2. + extents->x;
-	double center_y = extents->height / 2. + extents->y;
+	struct wlr_box extents;
+	wlr_output_layout_get_box(layout, NULL, &extents);
+	double center_x = extents.width / 2. + extents.x;
+	double center_y = extents.height / 2. + extents.y;
 
 	double dest_x = 0, dest_y = 0;
 	wlr_output_layout_closest_point(layout, NULL, center_x, center_y,
@@ -458,7 +408,12 @@ static struct wlr_output *wlr_output_layout_output_in_direction(
 		enum distance_selection_method distance_method) {
 	assert(reference);
 
-	struct wlr_box *ref_box = wlr_output_layout_get_box(layout, reference);
+	struct wlr_box ref_box;
+	wlr_output_layout_get_box(layout, reference, &ref_box);
+	if (wlr_box_empty(&ref_box)) {
+		// The output doesn't belong to the layout
+		return NULL;
+	}
 
 	double min_distance = (distance_method == NEAREST) ? DBL_MAX : DBL_MIN;
 	struct wlr_output *closest_output = NULL;
@@ -467,21 +422,23 @@ static struct wlr_output *wlr_output_layout_output_in_direction(
 		if (reference != NULL && reference == l_output->output) {
 			continue;
 		}
-		struct wlr_box *box = output_layout_output_get_box(l_output);
+
+		struct wlr_box box;
+		output_layout_output_get_box(l_output, &box);
 
 		bool match = false;
 		// test to make sure this output is in the given direction
 		if (direction & WLR_DIRECTION_LEFT) {
-			match = box->x + box->width <= ref_box->x || match;
+			match = box.x + box.width <= ref_box.x || match;
 		}
 		if (direction & WLR_DIRECTION_RIGHT) {
-			match = box->x >= ref_box->x + ref_box->width || match;
+			match = box.x >= ref_box.x + ref_box.width || match;
 		}
 		if (direction & WLR_DIRECTION_UP) {
-			match = box->y + box->height <= ref_box->y || match;
+			match = box.y + box.height <= ref_box.y || match;
 		}
 		if (direction & WLR_DIRECTION_DOWN) {
-			match = box->y >= ref_box->y + ref_box->height || match;
+			match = box.y >= ref_box.y + ref_box.height || match;
 		}
 		if (!match) {
 			continue;
